﻿using System.ComponentModel;
using System.Diagnostics;
using System.Runtime.InteropServices;

namespace UrsComPortExtensions
{
   /// <summary>
   /// Windows-Form mit den Ereignissen <see cref="DeviceAdded"/> und <see cref="DeviceRemoved"/>
   /// </summary>
   [ToolboxItem(false)]
   [DesignTimeVisible(false)]
   internal class UsbDetectionForm : Form
   {
      #region [ Ereignisdefinitionen ]
      /// <summary>
      /// Tritt ein, wenn ein Gerät aus dem System entfernt wurde.
      /// </summary>
      /// <param name="sender">Die Quelle des Ereignisses.</param>
      /// <param name="e">Ein Objekt, das informationen über das entfernte Gerät liefert.</param>
      internal delegate void DeviceRemovedEventHandler(object sender, DeviceChangedEventArgs e);
      /// <summary>
      /// Tritt ein, wenn ein Gerät aus dem System entfernt wurde.
      /// </summary>
      internal event DeviceRemovedEventHandler? DeviceRemoved;

      /// <summary>
      /// Tritt ein, wenn ein Gerät dem System hinzugefügt wurde.
      /// </summary>
      /// <param name="sender">Die Quelle des Ereignisses.</param>
      /// <param name="e">Ein Objekt, das informationen über das hinzugefügte Gerät liefert.</param>
      internal delegate void DeviceAddedEventHandler(object sender, DeviceChangedEventArgs e);
      /// <summary>
      /// Tritt ein, wenn ein Gerät dem System hinzugefügt wurde.
      /// </summary>
      internal event DeviceAddedEventHandler? DeviceAdded;
      #endregion

      #region [ Zugriff auf Windows-Funktionen ]
      private const int DBT_DEVICEARRIVAL = 0x8000;
      private const int DBT_DEVICEREMOVECOMPLETE = 0x8004;
      private const int WM_DEVICECHANGE = 0x219;
      private const int DBT_DEVTYP_DEVICEINTERFACE = 5;
      private static readonly Guid GuidDevinterfaceUSBDevice = new Guid("A5DCBF10-6530-11D2-901F-00C04FB951ED");
      private IntPtr notificationHandle;

      [DllImport("user32.dll", CharSet = CharSet.Auto, SetLastError = true)]
      private static extern IntPtr RegisterDeviceNotification(IntPtr recipient, IntPtr notificationFilter, int flags);

      [DllImport("user32.dll")]
      private static extern bool UnregisterDeviceNotification(IntPtr handle);

      [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Auto)]
      private partial struct DEV_BROADCAST_DEVICEINTERFACE
      {
         public int dbcc_size;
         public int dbcc_devicetype;
         public int dbcc_reserved;
         public Guid dbcc_classguid;
         [MarshalAs(UnmanagedType.ByValTStr, SizeConst = 255)]
         public string dbcc_name;
      }

      [StructLayout(LayoutKind.Sequential)]
      private partial struct DEV_BROADCAST_HDR
      {
         public int dbch_size;
         public int dbch_devicetype;
         public int dbch_reserved;
      }
      #endregion


      /// <summary>
      /// Initialisiert eine neue Instanz der <see cref="UsbDetectionForm"/>-Klasse.
      /// </summary>
      public UsbDetectionForm() : base()
      {
         RegisterUsbDeviceNotification();
      }

      #region [ Aktivieren/Deaktivieren ]
      /// <summary>
      /// Registiert das Fenster für den Empfang von WM_DEVICECHANGE-Nachrichten.
      /// </summary>
      protected void RegisterUsbDeviceNotification()
      {
         var dbi = new DEV_BROADCAST_DEVICEINTERFACE()
         {
            dbcc_devicetype = DBT_DEVTYP_DEVICEINTERFACE,
            dbcc_reserved = 0,
            dbcc_classguid = GuidDevinterfaceUSBDevice,
            dbcc_name = ""
         };
         dbi.dbcc_size = Marshal.SizeOf(dbi);
         var buffer = Marshal.AllocHGlobal(dbi.dbcc_size);
         Marshal.StructureToPtr(dbi, buffer, true);
         notificationHandle = RegisterDeviceNotification(Handle, buffer, 0);
      }

      /// <summary>
      /// Hebt die Registrierung wieder auf.
      /// </summary>
      protected void UnregisterUsbDeviceNotification()
      {
         UnregisterDeviceNotification(notificationHandle);
      }
      #endregion

      List<ComPortDescriptor> comPorts = ComPortDescriptor.GetPorts();


      /// <summary>
      /// Ermittelt die Device-ID aus den Nachrichtenparamtern.
      /// </summary>
      /// <param name="m">Die Nachrichtsdaten der DBT_DEVICEREMOVECOMPLETE- oder DBT_DEVICEARRIVAL-Nachricht.</param>
      /// <returns>Die ermittelte Device-ID im gleichen Format, wie im <b>ComDescriptor</b> abgelegt.</returns>
      private static string DeviceIDFromMesage(Message m)
      {
         DEV_BROADCAST_HDR hdr;
#pragma warning disable CS8605 // Unboxing eines möglichen NULL-Werts.
         hdr = (DEV_BROADCAST_HDR)Marshal.PtrToStructure(m.LParam, typeof(DEV_BROADCAST_HDR));

         if (hdr.dbch_devicetype == DBT_DEVTYP_DEVICEINTERFACE)
         {
            DEV_BROADCAST_DEVICEINTERFACE deviceInterface = (DEV_BROADCAST_DEVICEINTERFACE)Marshal.PtrToStructure(m.LParam, typeof(DEV_BROADCAST_DEVICEINTERFACE));
            string deviceName = new string(deviceInterface.dbcc_name);
            // Die DeviceID wird von der WMI-Abfrage und der Aulösung der Nachrichtenparameter in einem unterschiedlichen Format geliefert:

            // WMI:        "USB\VID_1A86&PID_7523\6&F9BED0F&0&3"
            // deviceName: "\\?\USB#VID_1A86&PID_7523#6&f9bed0f&0&3#{a5dcbf10-6530-11d2-901f-00c04fb951ed}"

            // Umformen auf das WMI-Format:
            String wmiName = "";
            wmiName = deviceName.Substring(4);
            int bracePos = wmiName.IndexOf("{");
            wmiName = wmiName.Substring(0, bracePos - 1);
            wmiName = wmiName.Replace("#", @"\");
            wmiName = wmiName.ToUpper();
            return wmiName;
         }
#pragma warning restore CS8605
         return "";
      }

      #region [ Ereignisse auslösen ]
      // Die Ereignisse dürfen nicht direkt in der WndProc ausgelöst werden.
      // Sie werden asynchron ausgelöst.
      private delegate void UsbDeviceAddedDelegate(string DeviceID);
      private void UsbDeviceAdded(string DeviceID)
      {
         String? portName = null;

         comPorts = ComPortDescriptor.GetPorts(); // Wir sind asynchron, die WMI ist bereits aktualisiert.
         foreach (var item in comPorts)
         {
            if(item.DeviceID == DeviceID)
            {
               portName =item.PortName;
               break;
            }
         }
         DeviceAdded?.Invoke(this, new DeviceChangedEventArgs(DeviceID, portName));
      }

      private delegate void UsbDeviceRemovedDelegate(string DeviceID);
      private void UsbDeviceRemoved(string DeviceID)
      {
         String? portName = null;
         foreach (var item in comPorts)
         {
            if (item.DeviceID == DeviceID)
            {
               portName = item.PortName;
               comPorts.Remove(item); // ComPort-Liste aktualisieren.
               break;
            }
         }
         DeviceRemoved?.Invoke(this, new DeviceChangedEventArgs(DeviceID, portName));
      }
      #endregion

      /// <summary>
      /// Windows-Funktion, die die Nachrichten DBT_DEVICE... auswertet.
      /// </summary>
      /// <param name="m"></param>
      [DebuggerStepThrough]
      protected override void WndProc(ref Message m)
      {
         base.WndProc(ref m);

         if (m.Msg == WM_DEVICECHANGE)
         {
            switch ((int)m.WParam)
            {
               case DBT_DEVICEREMOVECOMPLETE:
                  {
                     DeviceIDFromMesage(m);
                     BeginInvoke(new UsbDeviceRemovedDelegate(UsbDeviceRemoved), DeviceIDFromMesage(m));
                     break;
                  }
               case DBT_DEVICEARRIVAL:
                  {
                     DeviceIDFromMesage(m);
                     BeginInvoke(new UsbDeviceRemovedDelegate(UsbDeviceAdded), DeviceIDFromMesage(m));
                     break;
                  }
            }
         }
      }
   }
}
